# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://foglamp.dianomic.com
# FOGLAMP_END

""" Moving-Measure filter plugin
It can be used for the purpose of data-smoothing.
It supports the following weightage and measure
- Simple i.e. weightage of 1 & window size
    - Measure: Mean, Median, SEM, Min, Max, Sum, Std, Variance, Skew, Kurtosis
- Weighted which is weighted linearly with window size
    - Measure: LWMA
- Exponentially Weighted Moving Average
    - Measure: EMA or EWMA
"""


__author__ = "Mohit Rajput"
__copyright__ = "Copyright (c) 2023 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

PLUGIN_NAME = "moving_measure"

from foglamp.plugins.filter.moving_measure.mm_calc import _LOGGER

try:
    import filter_ingest
    from foglamp.plugins.filter.moving_measure.mm_calc import MovingMeasureExec
    _LOGGER.debug('Python packages in "{}" loaded successfully'.format(__file__))
except ImportError as e:
    _LOGGER.error(
        'Failed to load Python packages in "{}". {}'.format(
            __file__, str(e)
        )
    )

_grp1 = "Inputs"
_grp2 = "Strategy"
_grp3 = "Output"
_grp4 = _grp3  # 'Enable'

_DEFAULT_CONFIG = {
    "plugin": {
        "description": "Filter that calculates the Moving Measure and adds a new datapoint for a given asset datapoint pair",
        "type": "string",
        "default": PLUGIN_NAME,
        "readonly": "true",
    },
    # ----------------------- << Inputs >> ----------------------- #
    "in_asset": {
        "order": "1",
        "group": _grp1,
        "displayName": "Asset Name",
        "description": "Asset name of the data that will be the input to filter",
        "type": "string",
        "default": "input_asset",
    },
    "in_datapoint": {
        "order": "2",
        "group": _grp1,
        "displayName": "Datapoint Name",
        "description": "Asset attribute (datapoint) which will be the input to filter. Note, This must have numeric data.",
        "type": "string",
        "default": "input_datapoint",
    },
    # ----------------- << Calculation Config >> ----------------- #
    "cc_wt_strategy": {
        "order": "3",
        "group": _grp2,
        "displayName": "Smoothing Strategy",
        "description": "Weighted Strategy for data smoothing.",
        "type": "enumeration",
        "options": [
            "Simple (Non-Weighted)",
            "Linearly Weighted",
            "Exponential Weighted",
        ],
        "default": "Simple (Non-Weighted)",
    },
    "cc_window_size": {
        "order": "4",
        "group": _grp2,
        # note for future: issue with "cc_wt_strategy in [\"simple\", \"weighted\"]",
        "validity": 'cc_wt_strategy != "Exponential Weighted"',
        "displayName": "Window Size",
        "description": "Window size that is to be used for Moving Measure calculation",
        "type": "integer",
        "default": "10",
    },
    "cc_s_measure": {
        "order": "5",
        "group": _grp2,
        "validity": 'cc_wt_strategy == "Simple (Non-Weighted)"',
        "displayName": "Measure",
        "description": "Measure to calculate in the moving fixed size window",
        "type": "enumeration",
        "options": [
            "Mean",
            "Median",
            "Min",
            "Max",
            "Std",
            "SEM",
            "Sum",
            "Var",
            "Skew",
            "Kurtosis",
        ],
        "default": "Mean",
    },
    "cc_ema_parameter": {
        "order": "6",
        "group": _grp2,
        "validity": 'cc_wt_strategy == "Exponential Weighted"',
        "displayName": "EMA Parameter",
        "description": "The data-smoothing to be performed w.r.t. which parameter.",
        "type": "enumeration",
        "options": ["Alpha", "Span", "Center Of Mass", "Halflife"],
        "default": "Alpha",
    },
    "cc_ema_alpha": {
        "order": "7",
        "group": _grp2,
        "validity": 'cc_wt_strategy=="Exponential Weighted" && cc_ema_parameter=="Alpha"',
        "displayName": "EMA smoothing factor",
        "description": "Specify smoothing factor 0<alpha<=1",
        "type": "float",
        "minimum": "0.00001",
        "maximum": "1.0",
        "default": "0.2",
    },
    "cc_ema_span": {
        "order": "8",
        "group": _grp2,
        "validity": 'cc_wt_strategy=="Exponential Weighted" && cc_ema_parameter=="Span"',
        "displayName": "EMA Span",
        "description": "alpha = 2 / (span+1) for span >= 1 | Specify decay in terms of span. It corresponds to what is commonly called an - N-day Exponential Weighted moving average.",
        "type": "float",
        "minimum": "1.0",
        "default": "9",
    },
    "cc_ema_com": {
        "order": "9",
        "group": _grp2,
        "validity": 'cc_wt_strategy=="Exponential Weighted" && cc_ema_parameter=="Center Of Mass"',
        "displayName": "EMA Center Of Mass",
        "description": "alpha = 1 / (1+c) for com >= 0 | Specify decay in terms of center of mass. It has a more physical interpretation and can be thought of in terms of span",
        "type": "float",
        "minimum": "1.0",
        "default": "4",
    },
    "cc_ema_halflife": {
        "order": "10",
        "group": _grp2,
        "validity": 'cc_wt_strategy=="Exponential Weighted" && cc_ema_parameter=="Halflife"',
        "displayName": "EMA Halflife",
        "description": "alpha = 1 - exp**(log 0.5 / halflife) for halflife > 0 | Specify decay in terms of half-life. It is the period of time for the exponential weight to reduce to one half",
        "type": "float",
        "minimum": "0.0",
        "default": "3",
    },
    # ----------------------- << Output >> ----------------------- #
    "out_datapoint_suffix_auto": {
        "order": "11",
        "group": _grp3,
        "displayName": "Auto Generation",
        "description": 'Automatically decide the Output datapoint name based on selected measures "{{Input Datapoint}}{{Output Datapoint Suffix }}" 2nd part will be automatically added. ',
        "type": "boolean",
        "default": "true",
    },
    "out_datapoint_suffix": {
        "order": "12",
        "group": _grp3,
        "validity": 'out_datapoint_suffix_auto=="false"',
        "displayName": "Datapoint Suffix",
        "description": 'Output datapoint name will follow the pattern "{{Input Datapoint}}{{Output Datapoint Suffix }}" 2nd part is customizable. ',
        "type": "string",
        "default": "_moving_measure",
    },
    # ----------------------- << Enable >> ----------------------- #
    "enable": {
        "order": "13",
        "group": _grp4,
        "displayName": "Enabled",
        "description": "Enable Moving-Measure filter plugin",
        "type": "boolean",
        "default": "false",
    },
}


def plugin_info():
    """Return information about the plugin.

    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug(
        "Inside {} plugin_info(); Returning the plugin info details".format(
            PLUGIN_NAME
        )
    )
    return {
        "name": PLUGIN_NAME,
        "version": "3.0.0",
        "mode": "none",
        "type": "filter",
        "interface": "1.0",
        "config": _DEFAULT_CONFIG,
    }


def plugin_init(config, ingest_ref, callback):
    """Initialise the plugin.

    Args:
        config      : JSON configuration document for the Filter plugin
                      configuration category
        ingest_ref  : filter ingest reference
        callback    : filter callback
    Returns:
        handle      : JSON object to be used in future calls to the plugin
    Raises:
    """
    _LOGGER.debug("Inside {} plugin_init()".format(PLUGIN_NAME))
    _LOGGER.debug("\n\tconfig: " + str(config))

    handle = {}
    handle["mm"] = MovingMeasureExec(config)
    handle["ingestRef"] = ingest_ref
    handle["callback"] = callback

    _LOGGER.debug("\n\thandle created (updated config): " + str(handle))
    return handle


def plugin_reconfigure(handle, new_config):
    """Reconfigure the plugin.

    Args:
        handle     : handle returned by the plugin initialization call
        new_config : JSON object representing the new configuration category for
                     the category
    Returns:
        new_handle : new handle to be used in the future calls
    """
    _LOGGER.debug("Inside {} plugin_reconfigure()".format(PLUGIN_NAME))
    _LOGGER.debug(
        "Reconfiguring {}; \n\tOld Config: {} \n\tNew Config: {}".format(
            PLUGIN_NAME, handle, new_config
        )
    )

    new_handle = {}
    new_handle["mm"] = MovingMeasureExec(new_config)
    new_handle["ingestRef"] = handle["ingestRef"]
    new_handle["callback"] = handle["callback"]

    return new_handle


def plugin_shutdown(handle):
    """Shutdown the plugin doing required cleanup.

    Args:
        handle: handle returned by the plugin initialization call
    Returns:
    """
    _LOGGER.debug("Inside {} plugin_shutdown()".format(PLUGIN_NAME))

    handle["callback"] = None
    handle["ingestRef"] = None

    handle["mm"] = None

    _LOGGER.info("{} filter plugin shutdown completed.".format(PLUGIN_NAME))


def plugin_ingest(handle, data):
    """Modify readings data and pass it onward.

    Args:
        handle: handle returned by the plugin initialization call
        data: readings data
    """
    _LOGGER.debug("Inside {} plugin_ingest()".format(PLUGIN_NAME))
    _LOGGER.debug("handle:   " + str(handle))
    _LOGGER.debug("Incoming Data  :   " + str(type(data)) + " " + str(data))

    # calculated value for each reading
    process_data = None
    try:
        process_data = handle["mm"].calc_and_add_reading(data)
    except Exception as e:
        _LOGGER.error("Error in 'mm' Calculation : " + str(e))

    # Pass data onwards: if Filter is not enabled then process_data is None
    # (designed this way) in that case send unprocessed data onwords;
    filter_ingest.filter_ingest_callback(
        handle["callback"],
        handle["ingestRef"],
        data if process_data is None else process_data,
    )


    _LOGGER.debug("Outgoing Data  :   " + str(type(data)) + " " + str(data))
    _LOGGER.debug("{} filter ingestion performed".format(PLUGIN_NAME))
